import json
import secrets
import sys
from base64 import b64encode

import pytest
from pytest_django.asserts import assertContains, assertInHTML
from cryptojwt import JWS
from django.shortcuts import resolve_url
from django.utils.http import quote
from responses import matchers

from simple_openid_connect.data import TokenErrorResponse
from simple_openid_connect.integrations.django.apps import OpenidAppConfig


@pytest.mark.django_db
def test_directly_calling_login_endpoint(
    dyn_client,
    dummy_provider_config,
    dummy_provider_settings,
    response_mock,
    jwks,
    monkeypatch,
):
    # arrange
    settings = OpenidAppConfig.get_instance().safe_settings
    SECRET_CONSTANT = "42"
    monkeypatch.setattr(secrets, "token_urlsafe", lambda len: SECRET_CONSTANT)
    client_auth = b64encode(
        f"{settings.OPENID_CLIENT_ID}:{settings.OPENID_CLIENT_SECRET}".encode()
    ).decode()
    response_mock.get(
        url="https://provider.example.com/auth",
        match=[
            matchers.query_param_matcher(
                {
                    "client_id": settings.OPENID_CLIENT_ID,
                    "redirect_uri": settings.OPENID_BASE_URI
                    + resolve_url(settings.OPENID_REDIRECT_URI),
                    "response_type": "code",
                    "scope": settings.OPENID_SCOPE,
                    "nonce": SECRET_CONSTANT,
                    "state": SECRET_CONSTANT,
                }
            )
        ],
        status=302,
        headers={
            "Location": settings.OPENID_BASE_URI
            + resolve_url(settings.OPENID_REDIRECT_URI)
            + "?code=code.foobar123"
            + f"&state={SECRET_CONSTANT}"
        },
    )
    response_mock.post(
        url="https://provider.example.com/token",
        match=[
            matchers.urlencoded_params_matcher(
                {
                    "client_id": settings.OPENID_CLIENT_ID,
                    "code": "code.foobar123",
                    "grant_type": "authorization_code",
                    "redirect_uri": settings.OPENID_BASE_URI
                    + resolve_url(settings.OPENID_REDIRECT_URI),
                }
            ),
            matchers.header_matcher(
                {
                    "Authorization": f"Basic {client_auth}",
                }
            ),
        ],
        json={
            "access_token": "access_token.foobar123",
            "token_type": "Bearer",
            "id_token": JWS(
                json.dumps(
                    {
                        "iss": "https://provider.example.com",
                        "sub": "user1",
                        "aud": settings.OPENID_CLIENT_ID,
                        "iat": 0,
                        "exp": sys.maxsize,
                        "nonce": SECRET_CONSTANT,
                    }
                )
            ).sign_compact(jwks),
        },
    )

    # act
    response = dyn_client.get(
        "https://app.example.com" + resolve_url("simple_openid_connect:login"),
        follow=True,
    )

    # assert
    assert response.status_code == 200
    assert response.wsgi_request.path == resolve_url("default-after-login")
    assert response.content == b"default login redirect view"


@pytest.mark.django_db
def test_directly_accessing_protected_resource(
    dyn_client,
    dummy_provider_config,
    dummy_provider_settings,
    response_mock,
    jwks,
    monkeypatch,
):
    # arrange
    settings = OpenidAppConfig.get_instance().safe_settings
    SECRET_CONSTANT = "42"
    monkeypatch.setattr(secrets, "token_urlsafe", lambda len: SECRET_CONSTANT)
    client_auth = b64encode(
        f"{settings.OPENID_CLIENT_ID}:{settings.OPENID_CLIENT_SECRET}".encode()
    ).decode()
    response_mock.get(
        url="https://provider.example.com/auth",
        match=[
            matchers.query_param_matcher(
                {
                    "client_id": settings.OPENID_CLIENT_ID,
                    "redirect_uri": settings.OPENID_BASE_URI
                    + resolve_url(settings.OPENID_REDIRECT_URI),
                    "response_type": "code",
                    "scope": settings.OPENID_SCOPE,
                    "nonce": SECRET_CONSTANT,
                    "state": SECRET_CONSTANT,
                }
            )
        ],
        status=302,
        headers={
            "Location": settings.OPENID_BASE_URI
            + resolve_url(settings.OPENID_REDIRECT_URI)
            + "?code=code.foobar123"
            + f"&state={SECRET_CONSTANT}"
        },
    )
    response_mock.post(
        url="https://provider.example.com/token",
        match=[
            matchers.urlencoded_params_matcher(
                {
                    "client_id": settings.OPENID_CLIENT_ID,
                    "code": "code.foobar123",
                    "grant_type": "authorization_code",
                    "redirect_uri": settings.OPENID_BASE_URI
                    + resolve_url(settings.OPENID_REDIRECT_URI),
                }
            ),
            matchers.header_matcher(
                {
                    "Authorization": f"Basic {client_auth}",
                }
            ),
        ],
        json={
            "access_token": "access_token.foobar123",
            "token_type": "Bearer",
            "id_token": JWS(
                json.dumps(
                    {
                        "iss": "https://provider.example.com",
                        "sub": "user1",
                        "aud": settings.OPENID_CLIENT_ID,
                        "iat": 0,
                        "exp": sys.maxsize,
                        "nonce": SECRET_CONSTANT,
                    }
                )
            ).sign_compact(jwks),
        },
    )

    # act
    response = dyn_client.get(
        "https://app.example.com" + resolve_url("test-protected-view"), follow=True
    )

    # assert
    assert response.status_code == 200
    assert response.wsgi_request.path == resolve_url("test-protected-view")
    assert response.content == b"hello user user1"


@pytest.mark.django_db
def test_unsolicited_callback_csrf(
    dyn_client, dummy_provider_config, dummy_provider_settings, response_mock, jwks
):
    # arrange
    settings = OpenidAppConfig.get_instance().safe_settings

    # act
    # renders an error html page
    response = dyn_client.get(
        "https://app.example.com"
        + resolve_url(settings.OPENID_REDIRECT_URI)
        + "?code=code.foobar123"
    )

    # assert
    assertContains(response, "<h2>Invalid state</h2>", status_code=401, html=True)


@pytest.mark.django_db
def test_redirect_to_callback_with_error(
    monkeypatch,
    dyn_client,
    dummy_provider_config,
    dummy_provider_settings,
    response_mock,
):
    # arrange
    settings = OpenidAppConfig.get_instance().safe_settings
    SECRET_CONSTANT = "42"
    monkeypatch.setattr(secrets, "token_urlsafe", lambda len: SECRET_CONSTANT)
    response_mock.get(
        url="https://provider.example.com/auth",
        match=[
            matchers.query_param_matcher(
                {
                    "client_id": settings.OPENID_CLIENT_ID,
                    "redirect_uri": settings.OPENID_BASE_URI
                    + resolve_url(settings.OPENID_REDIRECT_URI),
                    "response_type": "code",
                    "scope": settings.OPENID_SCOPE,
                    "nonce": SECRET_CONSTANT,
                    "state": SECRET_CONSTANT,
                }
            )
        ],
        status=302,
        headers={
            "Location": settings.OPENID_BASE_URI
            + resolve_url(settings.OPENID_REDIRECT_URI)
            + f"?error=unauthorized_client&error_description={quote('Client is currently disabled')}&error_uri={quote('https://provider.example/error.html')}"
            + f"&state={SECRET_CONSTANT}"
        },
    )

    # act
    response = dyn_client.get(
        "https://app.example.com" + resolve_url("simple_openid_connect:login"),
        follow=True,
    )

    # assert
    assert response.status_code == 401
    response_html = response.content.decode("UTF-8")
    assertInHTML("<h1>Could not log you in</h1>", response_html)
    assertInHTML("<p>Client is currently disabled</p>", response_html)
    assertInHTML("<i>https://provider.example/error.html</i>", response_html)


@pytest.mark.django_db
def test_error_during_code_exchange(
    dyn_client,
    dummy_provider_config,
    dummy_provider_settings,
    response_mock,
    monkeypatch,
):
    # arrange
    settings = OpenidAppConfig.get_instance().safe_settings
    SECRET_CONSTANT = "42"
    monkeypatch.setattr(secrets, "token_urlsafe", lambda len: SECRET_CONSTANT)
    client_auth = b64encode(
        f"{settings.OPENID_CLIENT_ID}:{settings.OPENID_CLIENT_SECRET}".encode()
    ).decode()
    response_mock.get(
        url="https://provider.example.com/auth",
        match=[
            matchers.query_param_matcher(
                {
                    "client_id": settings.OPENID_CLIENT_ID,
                    "redirect_uri": settings.OPENID_BASE_URI
                    + resolve_url(settings.OPENID_REDIRECT_URI),
                    "response_type": "code",
                    "scope": settings.OPENID_SCOPE,
                    "nonce": SECRET_CONSTANT,
                    "state": SECRET_CONSTANT,
                }
            )
        ],
        status=302,
        headers={
            "Location": settings.OPENID_BASE_URI
            + resolve_url(settings.OPENID_REDIRECT_URI)
            + "?code=code.foobar123"
            + f"&state={SECRET_CONSTANT}"
        },
    )
    response_mock.post(
        url="https://provider.example.com/token",
        match=[
            matchers.urlencoded_params_matcher(
                {
                    "client_id": settings.OPENID_CLIENT_ID,
                    "code": "code.foobar123",
                    "grant_type": "authorization_code",
                    "redirect_uri": settings.OPENID_BASE_URI
                    + resolve_url(settings.OPENID_REDIRECT_URI),
                }
            ),
            matchers.header_matcher(
                {
                    "Authorization": f"Basic {client_auth}",
                }
            ),
        ],
        status=400,
        json=TokenErrorResponse(
            error=TokenErrorResponse.ErrorType.unsupported_grant_type,
            error_description="Grant is dummy disabled",
            error_uri="https://provider.example.com/error.html",
        ).model_dump(),
    )

    # act
    response = dyn_client.get(
        "https://app.example.com" + resolve_url("simple_openid_connect:login"),
        follow=True,
    )

    # assert
    assert response.status_code == 401
    response_html = response.content.decode("UTF-8")
    assertInHTML("<h1>Could not log you in</h1>", response_html)
    assertInHTML("<h2>unsupported_grant_type</h2>", response_html)
    assertInHTML("<p>Grant is dummy disabled</p>", response_html)
    assertInHTML("<i>https://provider.example.com/error.html</i>", response_html)



"""
JSON-Web-Key handling code
"""

from typing import List, cast

from cryptojwt import JWK, KeyBundle

from simple_openid_connect.exceptions import OpenidProtocolError


def fetch_jwks(jwks_uri: str) -> List[JWK]:
    """
    Fetch JSON web keys from the given jwks_uri.
    This uri is part of the provider configuration and used to validate responses and tokens sent by the provider.

    :raises OpenidProtocolError: If fetching the keys fails
    """
    bundle = KeyBundle(source=jwks_uri)
    keys = cast(List[JWK], bundle.keys())
    if not keys:
        raise OpenidProtocolError("Failed to fetch keys")
    return keys



# ruff: noqa: F405, F403        allow the * import
from .settings_base import *

# Database
# https://docs.djangoproject.com/en/3.2/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}

OPENID_BASE_URI = "https://app.example.com"
OPENID_ISSUER = "https://provider.example.com"
OPENID_CLIENT_ID = "test-client-id"
OPENID_CLIENT_SECRET = "test-client-secret"
OPENID_SCOPE = "openid profile email"



"""
Internal utilities
"""

from typing import Iterator

from simple_openid_connect.exceptions import ValidationError


def is_application_json(content_type: str) -> bool:
    """
    Whether the given content type is `application/json`.
    This is needed because mime types can contain additional options which are ignored here.
    """
    main_type = _parse_header(content_type)
    return main_type == "application/json"


def validate_that(condition: bool, msg: str) -> None:
    """
    Validate that the given condition is true, raising a ValidationError with the given message if it is not.

    This is implemented to write concise validating assertions.

    :raises ValidationError: if the condition is false
    """
    if not condition:
        raise ValidationError(msg)


def _parse_header(line: str) -> str:
    """Parse a Content-type like header.

    Return the main content-type.
    """
    # Adapted from https://github.com/python/cpython/blob/3.12/Lib/cgi.py#L238
    # by adding type hints and removing the parsing of additional mime type options.
    return _parseparam(";" + line).__next__()


def _parseparam(s: str) -> Iterator[str]:
    # Adapted from https://github.com/python/cpython/blob/3.12/Lib/cgi.py#L226
    # by adding type hints.
    while s[:1] == ";":
        s = s[1:]
        end = s.find(";")
        while end > 0 and (s.count('"', 0, end) - s.count('\\"', 0, end)) % 2:
            end = s.find(";", end + 1)
        if end < 0:
            end = len(s)
        f = s[:end]
        yield f.strip()
        s = s[end:]



from simple_openid_connect import pkce


def test_get_code_challenge():
    PKCE_PAIRS = [
        # (verifier, challenge) pairs
        (
            "X1mcNAU_lfu25acaJoCQSyo4YyZc1NwaxWw7tNL7mffK0AXCS-lNBtisryb2",
            "V_nSilRCu0pt3eR-cH7LWru-rYwlTr3J2143tiwloCA",
        ),
        (
            "9ok-JCOrX1OkmCFy8wGq2UBTGgFkeEg5IGffzgKNfDrpO-AOJ-83J9IhXOqj",
            "ET7UYFBGIHCdYh9bLyeLqWiKEr3t3JCxMRNacq3NCGk",
        ),
        (
            "zDv_Uz7QRoYRgIdieHac1UIlDNPPQb8qVIwgLzFu66kHo2g92UfkoU8Vi91z",
            "iWLceDSH1e3G7NJ5eieiqpk9IgXPTgSUd8ivdbRZKII",
        ),
    ]
    for verifier, challenge in PKCE_PAIRS:
        assert pkce.get_code_challenge(verifier) == challenge, (
            "get_code_challenge() produced an unexpected challenge"
        )


def test_gen_pair_is_actual_pair():
    for _ in range(100):
        verifier, challenge = pkce.generate_pkce_pair()
        assert challenge == pkce.get_code_challenge(verifier), (
            "pkce.generate_pkce_pair() returned a pair whose challenge cannot be reproduced from the verifier"
        )



#!/usr/bin/env python3
import argparse
import json
import logging
import urllib.parse
from collections import defaultdict
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler, HTTPServer
from secrets import token_urlsafe
from typing import Any, Dict
from typing import Optional  # noqa: F401

from simple_openid_connect.client import OpenidClient
from simple_openid_connect.data import (
    TokenSuccessResponse,
)

# initialize a session object (which is very primitive but works)
session = defaultdict(lambda: "")
client = None  # type: Optional[OpenidClient]


class RequestHandler(BaseHTTPRequestHandler):
    def do_GET(self) -> None:
        url = urllib.parse.urlparse(self.path)
        if url.path == "/":
            self._handle_initial()
        elif url.path.startswith("/callback"):
            self._handle_callback()
        else:
            self.return_redirect("/")

    def _handle_initial(self) -> None:
        global session
        global client
        assert client is not None

        # setup authorization code flow
        session["state"] = token_urlsafe(16)
        session["nonce"] = token_urlsafe(16)
        login_url = client.authorization_code_flow.start_authentication(
            state=session["state"], nonce=session["nonce"]
        )

        # send response
        self.return_redirect(login_url)

    def _handle_callback(self) -> None:
        global session
        global client
        assert client is not None

        # parse callback
        response = client.authorization_code_flow.handle_authentication_result(
            self.path, state=session["state"]
        )

        if isinstance(response, TokenSuccessResponse):
            access_token = response.access_token
            id_token = client.decode_id_token(response.id_token, nonce=session["nonce"])
            userinfo = client.fetch_userinfo(access_token)
            self.return_json_response(
                {
                    "token_response": response.model_dump(),
                    "id_token": id_token.model_dump(),
                    "userinfo": userinfo.model_dump(),
                }
            )

        else:
            self.return_json_response(response.model_dump())

    def return_redirect(self, to: str, code: int = HTTPStatus.FOUND) -> None:
        self.send_response(code)
        self.send_header("location", to)
        self.end_headers()

    def return_json_response(self, content: Dict[Any, Any]) -> None:
        self.send_response(HTTPStatus.OK)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps(content).encode("UTF-8"))


def main() -> None:
    argp = argparse.ArgumentParser(
        description="Script to retrieve access and refresh tokens from an OpenID-Connect identity provider and display them to the user. Useful for debugging purposes when requiring low-level access to these tokens."
    )
    argp.add_argument(
        "--issuer",
        required=True,
        help="OpenID_Connect issuer url",
    )
    argp.add_argument(
        "--client-id",
        required=True,
        help="OpenID-Connect client id",
    )
    argp.add_argument(
        "--client-secret", required=True, help="OpenID-Connect client secret"
    )
    argp.add_argument(
        "--scope",
        help="OpenID-Connect scopes to request. Default to just 'openid'. Can be given multiple times, once for each additional scope",
        action="append",
        default=["openid"],
    )
    argp.add_argument(
        "-p",
        "--port",
        type=int,
        default=8080,
        help="Port to bind this server to. Defaults to 8080",
    )
    argp.add_argument(
        "--host",
        default="127.0.0.1",
        help="IP address to bind this server to. Defaults to 127.0.0.1",
    )
    args = argp.parse_args()

    logging.basicConfig(level=logging.INFO)

    # initialize openid client
    global client
    own_base_url = f"http://{args.host}:{args.port}"
    client = OpenidClient.from_issuer_url(
        args.issuer,
        authentication_redirect_uri=own_base_url + "/callback",
        client_id=args.client_id,
        client_secret=args.client_secret,
        scope=" ".join(args.scope),
    )

    # serve a basic http server so that authorization code flow can be used
    with HTTPServer((args.host, args.port), RequestHandler) as server:
        print(f"Open http://{server.server_name}:{server.server_port}")
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            pass


if __name__ == "__main__":
    main()



"""
Django integration for :mod:`simple_openid_connect`.

Architecture
============

Integrating a new authentication mechanism into django involves a few missing parts.
Here's a quick overview over the most important classes involved:

.. code-block:: text

   ┌──────────────────────────────────┐   ┌──────────────────────────────────┐   ┌───────────────────┐
   │ @access_token_required decorator │   │ DRF AccessTokenAuthentication    │   │ LoginCallbackView │
   └────────────────┬─────────────────┘   └─────────────────┬────────────────┘   └─────────┬─────────┘
                    │                                       │                              │
                    └───────────────────┬───────────────────┘                              │
                                        ↓                                                  │
                 ┌──────────────────────┴─────────────────────┐                            │
                 │ UserMapper.handle_federated_access_token() │                            │
                 └──────────────────────┬─────────────────────┘                            │
                                        │                                                  │
                                        └──────────────────────────┬───────────────────────┘
                                                                   ↓
                                              ┌────────────────────┴───────────────────┐
                                              │ UserMapper.handle_federated_userinfo() │
                                              └────────────────────┬───────────────────┘
                                                                   ↓
                                                  ┌────────────────┴────────────────┐
                                                  │ UserMapper.automap_user_attrs() │
                                                  └─────────────────────────────────┘

- :class:`InitLoginView <simple_openid_connect.integrations.django.views.InitLoginView>` & :class:`LoginCallbackView <simple_openid_connect.integrations.django.views.LoginCallbackView>` are the backing views behind ``…/login/`` and ``…/login-callback/``.
  They implement authentication initiation with an OIDC provider and then handle the response to it in the callback.
- :class:`UserMapper <simple_openid_connect.integrations.django.user_mapping.UserMapper>` creates a local user account based on OIDC data and keeps it up to date.
  It is called every time a user authenticates, which is the case during interactive authentication as well as when a user passes an access token to e.g. an API route.
"""



"""
View-function decorators
"""

import logging
from functools import wraps
from http import HTTPStatus
from typing import Any, Callable, Optional, TypeVar, Union

from django.http import HttpRequest, HttpResponse, JsonResponse

from simple_openid_connect.exceptions import ValidationError
from simple_openid_connect.integrations.django.apps import OpenidAppConfig
from simple_openid_connect.utils import is_application_json

logger = logging.getLogger(__name__)

View_Return = TypeVar("View_Return", bound=HttpResponse)


def _invalid_token_response(request: HttpRequest) -> HttpResponse:
    if "Accept" in request.headers.keys() and is_application_json(
        request.headers["Accept"]
    ):
        return JsonResponse(
            status=HTTPStatus.UNAUTHORIZED,
            headers={"WWW-Authenticate": "Bearer"},
            data={
                "error": "invalid_token",
                "error_description": "the used access token is not valid or does not grant enough access",
            },
        )
    else:
        return HttpResponse(
            status=HTTPStatus.UNAUTHORIZED,
            content="the used access token is not valid or does not grant enough access",
            headers={"WWW-Authenticate": "Bearer"},
        )


def access_token_required(
    *, required_scopes: Optional[str] = None
) -> Callable[..., Union[HttpResponse, View_Return]]:
    """
    Decorator for views that checks that the request is authenticated using a valid access token, early-returning an
    appropriate http error response if necessary.

    :param required_scopes: Scopes to which the access token needs to have access.
        If not given, use the `settings.OPENID_SCOPE` value which defaults to "openid".

    :raises UnsupportedByProviderError: If the provider does not support token introspection.
    """
    if required_scopes is None:
        _required_scopes = OpenidAppConfig.get_instance().safe_settings.OPENID_SCOPE
    else:
        _required_scopes = required_scopes

    def actual_decorator(
        view_func: Callable[..., View_Return],
    ) -> Callable[..., View_Return]:
        @wraps(view_func)
        def wrapped_view(
            request: HttpRequest, *args: Any, **kwargs: Any
        ) -> Union[HttpResponse, View_Return]:
            # verify that an Authorization Header of type Bearer is present
            if "Authorization" not in request.headers.keys() or not request.headers[
                "Authorization"
            ].startswith("Bearer "):
                return HttpResponse(
                    status=HTTPStatus.UNAUTHORIZED,
                    headers={"WWW-Authenticate": "Bearer"},
                )

            oidc_client = OpenidAppConfig.get_instance().get_client(request)
            raw_token = request.headers["Authorization"].split(" ", 1)[1]

            try:
                (
                    request.user,
                    _,
                ) = OpenidAppConfig.get_instance().user_mapper.handle_federated_access_token(
                    raw_token, oidc_client, _required_scopes
                )
            except ValidationError:
                return _invalid_token_response(request)

            # execute the decorated view function
            return view_func(request, *args, **kwargs)

        return wrapped_view  # type: ignore

    return actual_decorator  # type: ignore



"""
Implementation of `PKCE <https://datatracker.ietf.org/doc/html/rfc7636>`_ code challenge and verifier generation.

Original code from `@RomeoDespres <https://github.com/RomeoDespres/>`_ in their `pkce repository <https://github.com/RomeoDespres/pkce/blob/b74b1864dc8a2018ca86566a5cfa2fd9fe751c4d/pkce/__init__.py>`_:


Examples
--------

>>> from simple_openid_connect import pkce
>>> code_verifier, code_challenge = pkce.generate_pkce_pair()

>>> from simple_openid_connect import pkce
>>> code_verifier = pkce.generate_code_verifier(length=128)
>>> code_challenge = pkce.get_code_challenge(code_verifier)
"""

import base64
import hashlib
import secrets
from typing import Tuple


def generate_code_verifier(length: int = 128) -> str:
    """
    Return a random PKCE-compliant code verifier.

    :param length: Code verifier length. Must be betwen 43 and 128.
    :raises ValueError: When `length` is not between 43 and 128.
    :returns: A url-safe string ready to be used as a code verifier.
    """
    if not 43 <= length <= 128:
        msg = "Parameter `length` must verify `43 <= length <= 128`."
        raise ValueError(msg)
    code_verifier = secrets.token_urlsafe(96)[:length]
    return code_verifier


def generate_pkce_pair(code_verifier_length: int = 128) -> Tuple[str, str]:
    """
    Return random PKCE-compliant code verifier and its corresponding code challenge.

    :param code_verifier_length: Length of the generated code verifier. Must be between 43 and 128.
    :raises ValueError: When `code_verifier_length` is not between 43 and 128.
    :returns code_verifier, code_challenge: A tuple containing the code verifier along with its corresponding challenge.
    """
    if not 43 <= code_verifier_length <= 128:
        msg = "Parameter `code_verifier_length` must verify "
        msg += "`43 <= code_verifier_length <= 128`."
        raise ValueError(msg)
    code_verifier = generate_code_verifier(code_verifier_length)
    code_challenge = get_code_challenge(code_verifier)
    return code_verifier, code_challenge


def get_code_challenge(code_verifier: str) -> str:
    """
    Generate the corresponding code challenge for a given verifier.

    :param code_verifier: The code verifier from which a challenge should be derived.
    :raises ValueError: When `code_verifier` is not bettween 43 and 128 long.
    :returns: The url-safe challenge string that corresponds to the given verifier.
    """
    if not 43 <= len(code_verifier) <= 128:
        msg = "Parameter `code_verifier` must verify "
        msg += "`43 <= len(code_verifier) <= 128`."
        raise ValueError(msg)
    hashed = hashlib.sha256(code_verifier.encode("ascii")).digest()
    encoded = base64.urlsafe_b64encode(hashed)
    code_challenge = encoded.decode("ascii")[:-1]
    return code_challenge



# ruff: noqa: F405, F403      allow the * import
from .settings_base import *

# Database
# https://docs.djangoproject.com/en/3.2/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}

OPENID_BASE_URI = "http://localhost:8000/"
OPENID_ISSUER = "http://localhost:8080/realms/dev"
OPENID_CLIENT_ID = "django_test_project"
OPENID_CLIENT_SECRET = "J2GslOfIEISn3XLXESIjU1X6SR9iaHB5"
OPENID_SCOPE = "openid profile email"



"""
View functions which handle openid authentication and their related callbacks
"""

import logging
import secrets
from datetime import datetime, timedelta, timezone
from http import HTTPStatus
from typing import Optional, Union, Any

from django.conf import settings
from django.contrib.auth import login, logout
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseRedirect,
)
from django.shortcuts import resolve_url
from django.template.response import TemplateResponse
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.cache import cache_control

from simple_openid_connect.data import (
    IdToken,
    RpInitiatedLogoutRequest,
    TokenSuccessResponse,
    TokenErrorResponse,
)
from simple_openid_connect.exceptions import ValidationError, AuthenticationFailedError
from simple_openid_connect.integrations.django.apps import OpenidAppConfig
from simple_openid_connect.integrations.django.models import OpenidSession

logger = logging.getLogger(__name__)


class InitLoginView(View):
    """
    The view which handles initiating a login.

    It essentially redirects the user agent to the Openid provider.
    """

    def get(self, request: HttpRequest) -> HttpResponse:
        logout(request)
        if "next" in request.GET.keys():
            request.session["login_redirect_url"] = request.GET["next"]

        # save the login state into the session to prevent CSRF attacks
        # ref: https://simple-openid-connect.readthedocs.io/en/stable/nonce_and_state.html
        state = secrets.token_urlsafe(32)
        request.session["openid_auth_state"] = state

        # save the time at which authentication was started
        request.session["openid_auth_start_time"] = datetime.now(
            tz=timezone.utc
        ).timestamp()

        # prevent replay attacks by generating and specifying a nonce
        # ref: https://simple-openid-connect.readthedocs.io/en/stable/nonce_and_state.html
        nonce = secrets.token_urlsafe(48)
        request.session["openid_auth_nonce"] = nonce

        # redirect the user-agent to the oidc provider
        client = OpenidAppConfig.get_instance().get_client(request)
        redirect = client.authorization_code_flow.start_authentication(
            state=state, nonce=nonce
        )
        return HttpResponseRedirect(redirect)


class LoginCallbackView(View):
    """
    The view which handles login callbacks.

    It handles an authentication response from the Openid provider that is encoded in the current url by either logging
    the user in or rendering the error.

    Error rendering can be customized by overwriting the template *simple_openid_connect/login_failed.html* which
    receives the context `token_response` of type :class:`TokenErrorResponse <simple_openid_connect.data.TokenErrorResponse>`.
    """

    def get(self, request: HttpRequest) -> HttpResponse:
        # Basic implementation flow is the following:
        # A number of steps are performed in order, each of which may return an HttpResponse object (that displays an error to the user).
        # If such an early response is detected, the whole view terminates early and returns that response.

        # do basic preliminary checks
        if (resp := self.check_auth_state(request)) is not None:
            return resp
        if (resp := self.check_login_timeout(request)) is not None:
            return resp

        # perform exchanges with the identity provider
        if isinstance(
            token_response := self.exchange_code_for_token(request), HttpResponse
        ):
            return token_response
        if isinstance(
            id_token := self.extract_id(request, token_response), HttpResponse
        ):
            return id_token

        # perform the actual login
        self.perform_login(request, token_response, id_token)

        # redirect to the next get parameter if present, otherwise to the configured default
        if "login_redirect_url" in request.session.keys():
            return HttpResponseRedirect(
                redirect_to=request.session["login_redirect_url"]
            )
        else:
            return HttpResponseRedirect(
                redirect_to=resolve_url(settings.LOGIN_REDIRECT_URL)
            )

    def render_error(
        self,
        request: HttpRequest,
        technical_msg: str,
        technical_description: str,
        **template_ctx: Any,
    ) -> HttpResponse:
        app_settings = OpenidAppConfig.get_instance()
        template_ctx.update(
            {
                "openid_settings": app_settings.safe_settings,
                "auth_redirect_url": app_settings.get_client(
                    request
                ).authentication_redirect_uri,
                "technical_msg": technical_msg,
                "technical_description": technical_description,
            }
        )
        return TemplateResponse(
            request,
            "simple_openid_connect/login_failed.html",
            template_ctx,
            status=HTTPStatus.UNAUTHORIZED,
        )

    def check_auth_state(self, request: HttpRequest) -> Optional[HttpResponse]:
        """
        Prevent CSRF attacks by verifying the requests state parameter

        ref: https://simple-openid-connect.readthedocs.io/en/stable/nonce_and_state.html
        """
        if request.session.get("openid_auth_state", None) is None:
            return self.render_error(
                request,
                "Invalid state",
                "It is only allowed to finish a login procedure if one has been started before. The state associated with your browser does not indicate that such a login process has been started before and you can therefore not be logged in right now. Please retry and ensure your browser does not delete cookies during naviation.",
            )

        return None

    def check_login_timeout(self, request: HttpRequest) -> Optional[HttpResponse]:
        """
        Don't allow login completion if the process was started too long ago
        """
        app_settings = OpenidAppConfig.get_instance().safe_settings

        if request.session.get("openid_auth_start_time", None) is None or (
            datetime.now(tz=timezone.utc)
            - datetime.fromtimestamp(
                request.session["openid_auth_start_time"], tz=timezone.utc
            )
        ) > timedelta(seconds=app_settings.OPENID_LOGIN_TIMEOUT):
            return self.render_error(
                request,
                "login process took too long",
                "For security reasons, loging in may only take a certain amount of time. Afterwards, the process must be restarted. This is done to prevent tricking a user into logging in when they did not request it right before.",
            )

        del request.session["openid_auth_start_time"]
        return None

    def exchange_code_for_token(
        self, request: HttpRequest
    ) -> Union[HttpResponse, TokenSuccessResponse]:
        """
        Exchange the code encoded in the current URL for an access token
        """
        client = OpenidAppConfig.get_instance().get_client(request)
        try:
            token_response = (
                client.authorization_code_flow.handle_authentication_result(
                    current_url=request.get_full_path(),
                    state=request.session["openid_auth_state"],
                )
            )
        except AuthenticationFailedError as e:
            return self.render_error(
                request,
                "The Identity-Provider ran into an error during the authentication process",
                "This can happen and is entirely in the discretion of the identity provider. See the additional message below for more info.",
                auth_failed_error=e.error,
            )
        if isinstance(token_response, TokenErrorResponse):
            return self.render_error(
                request,
                str(token_response.error),
                "This application was not able to exchange the *code* it got during the users navigation for proper access tokens. See the additional message below for more info.",
                token_error_response=token_response,
            )
        else:
            return token_response

    def extract_id(
        self, request: HttpRequest, token_response: TokenSuccessResponse
    ) -> Union[HttpResponse, IdToken]:
        """
        Extract the validated ID token from the given token response
        """
        client = OpenidAppConfig.get_instance().get_client(request)
        id_token = IdToken.parse_jwt(token_response.id_token, client.provider_keys)
        try:
            id_token.validate_extern(
                client.provider_config.issuer,
                client.client_auth.client_id,
                nonce=request.session["openid_auth_nonce"],
            )
            return id_token
        except ValidationError as e:
            return self.render_error(
                request,
                "Invalid ID-Token",
                "This application validates ID-Tokens (user information received from the identity provider) based on the providers cryptographic keys and based on certain properties during authentication. This validation failed for the reason indicated below.",
                id_token_validation_error=e,
            )

    def perform_login(
        self,
        request: HttpRequest,
        token_response: TokenSuccessResponse,
        id_token: IdToken,
    ) -> Any:
        """
        Perform internal user object maintenance and authentication after having performed all external communication

        This handles federated user information (create a new user if necessary or update local info) and logs the user in.
        """
        user = OpenidAppConfig.get_instance().user_mapper.handle_federated_userinfo(
            id_token
        )
        openid_session = user.openid.update_session(token_response, id_token)
        request.session["openid_session"] = openid_session.id
        login(request, user, backend=settings.AUTHENTICATION_BACKENDS[0])


class LogoutView(View):
    """
    The view which handles logging a user out.
    """

    def get(self, request: HttpRequest) -> HttpResponse:
        session_id = request.session.get("openid_session")
        logout(request)
        client = OpenidAppConfig.get_instance().get_client(request)

        if settings.LOGOUT_REDIRECT_URL is not None:
            openid_session = (
                OpenidSession.objects.get(id=session_id) if session_id else None
            )

            logout_request = RpInitiatedLogoutRequest(
                post_logout_redirect_uri=request.build_absolute_uri(
                    resolve_url(settings.LOGOUT_REDIRECT_URL)
                )
            )
            if openid_session is not None and openid_session.raw_id_token is not None:
                logout_request.id_token_hint = openid_session.raw_id_token
            else:
                logout_request.client_id = client.client_auth.client_id
        else:
            logout_request = None

        return HttpResponseRedirect(client.initiate_logout(logout_request))


class FrontChannelLogoutNotificationView(View):
    """
    A view which handles Openid front-channel logout notifications by logging out the current session
    """

    @method_decorator(cache_control(no_store=True))
    def get(self, request: HttpRequest) -> HttpResponse:
        logout(request)
        return HttpResponse(status=200)
